import copy
import numpy as np

import simdb.run as sdb
from problems import h_problem
from lrb_operator_projection import LRBOperatorProjection
from localizer import NumpyLocalizer
from partitioner import build_subspaces, partition_any_grid

from my_discretize_elliptic_cg import discretize_elliptic_cg
#from pymor.discretizers.elliptic import discretize_elliptic_cg
from pymor.vectorarrays.numpy import NumpyVectorSpace
from pymor.discretizations.basic import StationaryDiscretization
from pymor.domaindescriptions.boundarytypes import BoundaryType
from pymor.operators.constructions import induced_norm, VectorFunctional
from pymor.algorithms.basisextension import gram_schmidt_basis_extension

from pymor.operators.numpy import NumpyMatrixOperator
from pymor.vectorarrays.numpy import NumpyVectorArray

import scipy.sparse
from mybmat import mybmat
scipy.sparse.bmat = mybmat




experimentname = "globally_coupled_convergence"
resolution = 200
coarse_grid_resolution = 10
problem = "h"
contrast=1e5

sdb.new_dataset(experimentname,
                problem=problem,
                contrast=contrast,
                resolution=resolution,
                coarse_grid_resolution=coarse_grid_resolution)

if problem == "h":
    p = h_problem(contrast=contrast)
else:
    raise "wrong problem"

diameter = 1./resolution
d, data = discretize_elliptic_cg(p, diameter=diameter)

# hack for more numerical stability
dirichlet_dofs = data['boundary_info'].dirichlet_boundaries(2)
for op in d.operator.operators:
    op.assemble()._matrix[dirichlet_dofs, dirichlet_dofs] *= 1e5
    d.rhs.assemble()._matrix[:, dirichlet_dofs] *= 1e5


op = d.operator.assemble()
rhs = d.rhs.assemble()
grid = data["grid"]

subspaces, subspaces_per_codim = build_subspaces(
    *partition_any_grid(grid, num_intervals=(coarse_grid_resolution, coarse_grid_resolution))
     )
localizer = NumpyLocalizer(d.solution_space, subspaces['dofs'])

spaces = [subspaces[s_id]["env"] for s_id in subspaces_per_codim[2]]
assert len(spaces) == 81

bases = {}
empty_bases = {}
for space in spaces:
    space_size = len(localizer.join_spaces(space))
    bases[space] = NumpyVectorSpace(space_size).empty()
    empty_bases[space] = NumpyVectorSpace(space_size).empty()


operator_reductor = LRBOperatorProjection(op, rhs, localizer, spaces, bases, spaces, bases)

full_energy_product = copy.deepcopy(d.operator.assemble())
d_indices = np.nonzero(data["boundary_info"].mask(BoundaryType("dirichlet"), 2))[0]
if len(d_indices):
    full_energy_product._matrix[d_indices,:] *= 0
    full_energy_product._matrix[:,d_indices] *= 0
    full_energy_product._matrix[d_indices, d_indices] = 1
    full_energy_product._matrix.eliminate_zeros()

energy_norm = induced_norm(full_energy_product)
def energy_riesz(vec):
    return full_energy_product.apply_inverse(vec)

energy_products = {}
for space in spaces:
    energy_products[space] = localizer.localize_operator(full_energy_product, space, space)

u = d.solve()

debugout = open("debugout.txt", "w")

extensions = NumpyVectorSpace(op.source.dim).empty()

for iteration in range(700):
    rop = operator_reductor.get_reduced_operator()
    rrhs = operator_reductor.get_reduced_rhs()
    rd = StationaryDiscretization(rop,rrhs,cache_region=None)
    ru = operator_reductor.reconstruct_source(rd.solve())
    error = u - ru
    residual = rhs.as_vector() - op.apply(ru)
    # error is riesz representative of residual in energy norm
    #assert np.abs((energy_norm(error) - energy_norm(energy_riesz(residual))) / np.abs(energy_norm(error))) < 1e-5

    max_shift = 0.
    max_shift_space = None
    for space in spaces:
        operator_reductor.set_source_basis(space, None)
        operator_reductor.set_range_basis(space, None)
        lp_op = operator_reductor.get_reduced_operator()
        lp_rhs = operator_reductor.get_reduced_rhs()
        lp_d = StationaryDiscretization(lp_op, lp_rhs, cache_region=None)
        lp_u = operator_reductor.reconstruct_source(lp_d.solve())
        lp_shift = ru - lp_u
        lp_shift_norm = energy_norm(lp_shift)[0]

        operator_reductor.set_source_basis(space, bases[space])
        operator_reductor.set_range_basis(space, bases[space])
        #print("error in {} is {}".format(space, energy_norm(lp_error)))
        if lp_shift_norm > max_shift:
            max_shift = lp_shift_norm
            max_shift_space = space

    e_space = max_shift_space
    print("einriching{}".format(e_space))

    operator_reductor.set_source_basis(e_space, None)
    operator_reductor.set_range_basis(e_space, None)
    e_operator = operator_reductor.get_reduced_operator()
    e_rhs = operator_reductor.get_reduced_rhs()

    e_d = StationaryDiscretization(e_operator, e_rhs, cache_region=None)
    u_l = operator_reductor.reconstruct_source_part(e_d.solve(), e_space)

    norm_u_l = induced_norm(energy_products[e_space])(u_l)[0]

    e_basis = bases[e_space]
    try:
        gram_schmidt_basis_extension(e_basis, u_l, product=energy_products[e_space], copy_basis=False)
    except:
        break

    operator_reductor.set_source_basis(e_space, e_basis)
    operator_reductor.set_range_basis(e_space, e_basis)

    sdb.append_values(
        error_energy_norm=energy_norm(error)[0],
        error_h1_norm=d.h1_norm(error)[0],
        error_l2_norm=d.l2_norm(error)[0],
        residual_dual_norm=energy_norm(energy_riesz(residual))[0],
        norm_u_l=norm_u_l,
        )
        
    print("Fehler: {}".format(energy_norm(error)))
    debugout.write("{}\n".format(energy_norm(error)[0]))
    debugout.flush()
    sdb.flush()
